-- Copyright  2008, 2009 Xianghar  <xian@zron.de>
-- All Rights Reserved.
-- This code is not to be modified or distributed without written permission by the author.

TotemTimers = {}
TotemTimers.numtimers = 0
TotemTimers.__index = TotemTimers
TotemTimers.timers = {}

local Timers = TotemTimers.timers

local function FormatTime(frame, sec, format)
	local seconds = floor(sec)

 	if format == "blizz" then
		frame:SetFormattedText(SecondsToTimeAbbrev(sec))
    elseif format == "sec" then
        seconds = floor(sec+0.5)
        if seconds == 0 then seconds = "" end
        frame:SetFormattedText(tostring(seconds))
	else
		if(seconds <= 0) then
			frame:SetText("")
		elseif seconds < 600 then
			local d, h, m, s = ChatFrame_TimeBreakDown(seconds)
			frame:SetFormattedText("%01d:%02d", m, s)
		elseif(seconds < 3600) then
			local d, h, m, s = ChatFrame_TimeBreakDown(seconds);
			frame:SetFormattedText("%02d:%02d", m, s)
		else
			frame:SetText("1 hr+")
		end
	end
end


function TotemTimers_UpdateTimers(elapsed) 
	for _,v in pairs(Timers) do
		if v.active then v:update(elapsed) end
	end
end

function TotemTimers:new(numtimers)
	local self = {}
	setmetatable(self, TotemTimers)
	TotemTimers.numtimers = TotemTimers.numtimers + 1
    self.nr = TotemTimers.numtimers
	self.active = false
	self.button = CreateFrame("Button", "TotemTimers_Timer"..TotemTimers.numtimers, UIParent, "TotemTimersTemplate")
	self.button:SetPoint("CENTER", UIParent, "CENTER")
	self.button.timer = self
	
	--for rActionButtonStyler
	self.button.action = 0 
	self.button.SetCheckedTexture = function() end
	if not IsAddOnLoaded("rActionButtonStyler") then
		self.button:SetNormalTexture(nil)
	else
		ActionButton_Update(self.button)
	end
	--
	
	self.button.element = TotemTimers.numtimers
	
	self.numtimers = numtimers
	self.timers = {}
    for i=1, numtimers do self.timers[i] = 0 end
	self.timerbars = {}
	self.button.icons = {}
	self.button.flash = {}
	for i=1,numtimers do
		self.timers[i] = 0
		self.timerbars[i] = CreateFrame("StatusBar", "TotemTimers_TimerBar"..TotemTimers.numtimers.."_"..i, self.button, "TotemTimersTimerBarTemplate")
        self.timerbars[i].background = getglobal("TotemTimers_TimerBar"..TotemTimers.numtimers.."_"..i.."Background")
        self.timerbars[i].time = getglobal("TotemTimers_TimerBar"..TotemTimers.numtimers.."_"..i.."Time")
        self.timerbars[i].icon = getglobal("TotemTimers_TimerBar"..TotemTimers.numtimers.."_"..i.."Icon")
        self.timerbars[i]:SetPoint("TOP", self.button, "BOTTOM")
		self.button.icons[i] = getglobal("TotemTimers_Timer"..TotemTimers.numtimers.."Icon"..((i>1) and i or ""))
		self.button.flash[i] = getglobal("TotemTimers_Timer"..TotemTimers.numtimers.."Flash"..((i>1) and i or ""))
		self:SetAlpha(self.button.icons[i], 0.4)
	end
	self.button.icons[1]:Show()
	self.timerBarPos = "BOTTOM"
	self.timeSpacing = 0
	self.spacing = 5
	self.timeStyle = "mm:ss"
	self.warningMsgs = {}
	self.expirationMsgs = {}
	self.earlyExpirationMsgs = {}
	self.events = {}
	self.button.count = getglobal("TotemTimers_Timer"..TotemTimers.numtimers.."Count")
	self.button.cooldown = getglobal("TotemTimers_Timer"..TotemTimers.numtimers.."Cooldown")
	self.button.miniIcon = getglobal("TotemTimers_Timer"..TotemTimers.numtimers.."MiniIcon")
    self.button.miniIconFrame = getglobal("TotemTimers_Timer"..TotemTimers.numtimers.."Mini")
    self.button.bar = getglobal("TotemTimers_Timer"..TotemTimers.numtimers.."Bar")
    self.button.hotkey = getglobal("TotemTimers_Timer"..TotemTimers.numtimers.."HotKey")
    self.button.playerdot = getglobal("TotemTimers_Timer"..TotemTimers.numtimers.."PlayerBuff")
    self.button.buffCount = getglobal("TotemTimers_Timer"..TotemTimers.numtimers.."BuffCount")
	
	self.anchors = {}
	self.anchorchilds = {}
	
	self.button:SetScript("OnDragStart", TotemTimers.StartMoving)
	self.button:SetScript("OnDragStop", TotemTimers.StopMoving)
	self.button:RegisterForDrag("LeftButton")
	
	table.insert(Timers, self)
    self.button:Hide()
	return self
end

-- timer functions
local IsSpellInRange = IsSpellInRange
local IsUsableSpell = IsUsableSpell

function TotemTimers:update(elapsed)
	local timers = self.timers
	for i=1,self.numtimers do
		if timers[i] > 0 then
			timers[i] = timers[i] - elapsed
			if timers[i] <= 0 then
				self:stop(i) 
			else
				if timers[i]<10 and self.warningMsgs[i] then
					self:Msg(self.warningMsgs[i], "warning")
					self.warningMsgs[i] = nil
				end
                if timers[i] >= 600 then
                    FormatTime(self.timerbars[i].time, timers[i], "blizz")
                else
                    FormatTime(self.timerbars[i].time, timers[i], self.timeStyle)
                end
				if self.visibleTimerBars then
					self.timerbars[i]:SetValue(timers[i])
				end
				if timers[i] <= 10 then
					self:Flash(i)
				end
			end
		end
	end
    if self.ProcFlash and timers[1] > 0 and self.bar then
        self.button.bar:SetValue(self.bar - mod(timers[1], self.bar))
    end
    self.button.icons[1]:SetVertexColor(1,1,1)
    self.button.miniIcon:SetVertexColor(1,1,1)
    if self.RangeCheck then
        if TotemTimers_Spells[self.RangeCheck] then
            if IsSpellInRange(TotemTimers_Spells[self.RangeCheck]) == 0 then
                self.button.icons[1]:SetVertexColor(1,0,0)
            end
        end
    end
    if self.ManaCheck then
        if TotemTimers_Spells[self.ManaCheck] then
            local _,nomana = IsUsableSpell(TotemTimers_Spells[self.ManaCheck])
            if nomana then
                if self.ManaCheckMini then
                    self.button.miniIcon:SetVertexColor(0.5,0.5,1)
                else
                    self.button.icons[1]:SetVertexColor(0.5,0.5,1)
                end
            end
        end
    end
end

function TotemTimers:activate()
	self.active = true
	for _,event in pairs(self.events) do
		self.button:RegisterEvent(event)
	end
	self.button:Show()
end

function TotemTimers:deactivate()
	self.active = false
	for _,event in pairs(self.events) do
		self.button:UnregisterEvent(event)
	end
	self.button:Hide()
end

function TotemTimers:start(timer, time, duration)
	duration = duration or time
	local timerbar = self.timerbars[timer]
	self.timers[timer] = time
	FormatTime(timerbar.time, time, self.timeStyle)
	timerbar:SetMinMaxValues(0, duration)
    self:ShowTimerBar(timer)
	if self.visibleTimerBars then
		timerbar:SetValue(time)
	end
	if not self.dontAlpha then self:SetAlpha(self.button.icons[timer], 1) end
    if self.reverseAlpha then self:SetAlpha(self.button.icons[timer], 0.4) end
	if self.showCooldown then
		CooldownFrame_SetTimer(self.button.cooldown, GetTime(), duration, 1)
	end
    if self.button.flash[timer] then self.button.flash[timer]:Hide() end
    self:SetTimerBarPos(self.timerBarPos, true)
end

function TotemTimers:stop(timer)
	local timerbar = self.timerbars[timer]
	if self.timers[timer]>1 then
		if self.earlyExpirationMsgs[timer] then
			self:Msg(self.earlyExpirationMsgs[timer], "expiration")
			self.earlyExpirationMsgs[timer] = nil
		end
	elseif self.expirationMsgs[timer] then
		self:Msg(self.expirationMsgs[timer], "expiration")
		self.expirationMsgs[timer] = nil
	end
	self.timers[timer] = 0
	timerbar.time:SetText("")
	if self.visibleTimerBars then		
		timerbar:SetValue(0)
	end
    self:HideTimerBar(timer)
    if not self.dontAlpha then self:SetAlpha(self.button.icons[timer], 0.4) end
    if self.reverseAlpha and InCombatLockdown() then self:SetAlpha(self.button.icons[timer],1) end
	if self.button.flash[timer] then self.button.flash[timer]:Hide() end
    if self.ProcFlash and timer == 1 then 
        self.bar = nil
        self.button.bar:SetValue(0)
    end
    self.button.playerdot:Hide()
    self:SetTimerBarPos(self.timerBarPos, true)
end


function TotemTimers:StartMoving()
    if self.timer.locked then return end
	if self.anchorframe and not self.timer.savePos then
		self.anchorframe:StartMoving()
	else
		self:StartMoving()
	end
end

function TotemTimers:StopMoving()
	if self.anchorframe and not self.timer.savePos then
		self.anchorframe:StopMovingOrSizing()
	else
		self:StopMovingOrSizing()
	end
    if self.timer.savePos and self:GetNumPoints()>0 then
        --[[if not TotemTimers_Settings.TimerPositions[self.timer.nr] then
            TotemTimers_Settings.TimerPositions[self.timer.nr] = {}
        end]]
		local pos = {self:GetPoint(1)}
		if not pos[1] then pos = nil
        elseif pos[2] then
            pos[2] = pos[2]:GetName()
        end
		TotemTimers_Settings.TimerPositions[self.timer.nr] = pos
    end
    self.timer:SetTimerBarPos(self.timer.timerBarPos, true)
end

function TotemTimers:SetAlpha(icon, alpha)
    if icon then
        icon:SetAlpha(alpha)
    end
end

function TotemTimers:HideTimerBar(nr)
	self.timerbars[nr].background:Hide()
    self.timerbars[nr].background:SetValue(0)
    ---self.timerbars[nr].icon:Hide()
    self.timerbars[nr]:Hide()
end

function TotemTimers:ShowTimerBar(nr)
    self.timerbars[nr]:Show()
	if self.visibleTimerBars then
        self.timerbars[nr].background:Show()
        self.timerbars[nr].background:SetValue(1)
    end
end

function TotemTimers:Flash(nr)
    if self.dontFlash or (nr > 1 and not self.flashall) or nr > 2 then return end
	local flash = self.FlashRed and self.button.flash[nr] or self.button.icons[nr]
    
	flash:Show()
	flash:SetAlpha(BuffFrame.BuffAlphaValue)
end


-- display functions

function TotemTimers:SetTimerBarPos(side, notReanchor)
	self.timerBarPos = side
	
	local timerbars = self.timerbars
	local numtimers = self.numtimers
		
	for i=1, numtimers do
		timerbars[i]:ClearAllPoints()
        timerbars[i].icon:ClearAllPoints()
	end
    if side == "RIGHT" then
        for i=2,numtimers do
            timerbars[i].icon:SetPoint("LEFT", timerbars[i], "RIGHT", -4, 0)
        end
    else
        for i=2,numtimers do
            timerbars[i].icon:SetPoint("RIGHT", timerbars[i], "LEFT", 4, 0)
        end
    end
    local activetimers = 1
	if side == "LEFT" then
        local lastactive = 1
		for i=2,numtimers do
            timerbars[i]:SetPoint("TOP", timerbars[lastactive], "BOTTOM") -- set all, but only active ones will be seen so they can lay atop one another
            if self.timers[i] > 0 then
                lastactive = i
                activetimers = activetimers + 1
            end
		end
		timerbars[1]:SetPoint("RIGHT", self.button, "LEFT", -self.timeSpacing, timerbars[1]:GetHeight()*timerbars[1]:GetEffectiveScale()/2*(activetimers-1))
	elseif side == "RIGHT" then
        local lastactive = 1
		for i=2,numtimers do
            timerbars[i]:SetPoint("TOP", timerbars[lastactive], "BOTTOM")
            if self.timers[i] > 0 then
                lastactive = i
                activetimers = activetimers + 1
            end
		end
		timerbars[1]:SetPoint("LEFT", self.button, "RIGHT", self.timeSpacing, timerbars[1]:GetHeight()*timerbars[1]:GetEffectiveScale()/2*(activetimers-1))
	elseif side == "TOP" then
        if TotemTimers_Settings.ShowCooldowns and self.nr < 5 and (TotemTimers_Settings.Arrange == "vertical" or TotemTimers_Settings.Arrange == "box") then
            timerbars[1]:SetPoint("BOTTOM", self.button, "TOP", 0, self.timeSpacing)
            local active = 0
            local firstactive = 0
            for i = 2, numtimers do
                if self.timers[i] > 0 then
                    active = active + 1
                    if firstactive == 0 then firstactive = i end
                end
            end
            if active > 0 then
                local left = self.button:GetLeft() < WorldFrame:GetWidth()/2
                if TotemTimers_Settings.Arrange == "box" and self.actnr then
                    left = self.actnr == 2 or self.actnr == 4
                end
                for i=2, numtimers do
                    timerbars[i].icon:ClearAllPoints()
                    if left then
                        timerbars[i].icon:SetPoint("LEFT", timerbars[i], "RIGHT", -4, 0)
                    else
                        timerbars[i].icon:SetPoint("RIGHT", timerbars[i], "LEFT", 4, 0)
                    end
                end
                if left then
                    timerbars[firstactive]:SetPoint("LEFT", self.button, "RIGHT", 
                        self.timeSpacing, timerbars[firstactive]:GetHeight()*timerbars[firstactive]:GetEffectiveScale()/2*(active-1))
                else
                    timerbars[firstactive]:SetPoint("RIGHT", self.button, "LEFT",
                        -self.timeSpacing, timerbars[firstactive]:GetHeight()*timerbars[firstactive]:GetEffectiveScale()/2*(active-1))
                end
                local last = firstactive
                for i = firstactive+1, numtimers do
                    timerbars[i]:SetPoint("TOP", timerbars[last], "BOTTOM")
                    if self.timers[i] > 0 then                        
                        last = i
                    end
                end
                for i = 2, firstactive-1 do
                    timerbars[i]:SetPoint("TOP", self.button, "BOTTOM")
                end
            end
        else
            local lastactive = 1
    		timerbars[1]:SetPoint("BOTTOM", self.button, "TOP", 0, self.timeSpacing)
    		for i=2, numtimers do
                timerbars[i]:SetPoint("BOTTOM", timerbars[lastactive], "TOP")
                if self.timers[i] > 0 then
                    lastactive = i
                end
    		end
        end
	elseif side == "BOTTOM" then
        if TotemTimers_Settings.ShowCooldowns and self.nr < 5 and (TotemTimers_Settings.Arrange == "vertical" or TotemTimers_Settings.Arrange == "box") then
            timerbars[1]:SetPoint("TOP", self.button, "BOTTOM", 0, -self.timeSpacing)
            local active = 0
            local firstactive = 0
            for i = 2, numtimers do
                if self.timers[i] > 0 then
                    active = active + 1
                    if firstactive == 0 then firstactive = i end
                end
            end
            if active > 0 then
                local left = self.button:GetLeft() < WorldFrame:GetWidth()/2
                if TotemTimers_Settings.Arrange == "box" and self.actnr then
                    left = self.actnr == 2 or self.actnr == 4
                end
                for i=2, numtimers do
                    timerbars[i].icon:ClearAllPoints()
                    if left then
                        timerbars[i].icon:SetPoint("LEFT", timerbars[i], "RIGHT", -4, 0)
                    else
                        timerbars[i].icon:SetPoint("RIGHT", timerbars[i], "LEFT", 4, 0)
                    end
                end
                if left then
                    timerbars[firstactive]:SetPoint("LEFT", self.button, "RIGHT", 
                        self.timeSpacing, timerbars[firstactive]:GetHeight()*timerbars[firstactive]:GetEffectiveScale()/2*(active-1))
                else
                    timerbars[firstactive]:SetPoint("RIGHT", self.button, "LEFT",
                        -self.timeSpacing, timerbars[firstactive]:GetHeight()*timerbars[firstactive]:GetEffectiveScale()/2*(active-1))
                end
                local last = firstactive
                for i = firstactive+1, numtimers do
                    timerbars[i]:SetPoint("TOP", timerbars[last], "BOTTOM")
                    if self.timers[i] > 0 then
                        last = i
                    end
                end
                for i = 2, firstactive-1 do
                    timerbars[i]:SetPoint("TOP", self.button, "BOTTOM")
                end
             end
        else
            local lastactive = 1
    		timerbars[1]:SetPoint("TOP", self.button, "BOTTOM", 0, -self.timeSpacing)
    		for i=2, numtimers do
                timerbars[i]:SetPoint("TOP", timerbars[lastactive], "BOTTOM")
                if self.timers[i] > 0 then
                    lastactive = i
                end
    		end
        end
	end
	if not InCombatLockdown() and not notReanchor then self:Reanchor() end
end

function TotemTimers:GetBorder(side)
	local timerBarPos = self.timerBarPos
	if side == "TOP" and timerBarPos == "TOP" or side == "BOTTOM" and timerBarPos == "BOTTOM" then
        local height = self.timerbars[1]:GetHeight()*self.timerbars[1]:GetEffectiveScale()
        if self.nr > 5 or TotemTimers_Settings.Arrange == "horizontal" then
            height = height * self.numtimers
        end
		return self.timeSpacing + height
	elseif (side == "LEFT" and timerBarPos == "LEFT" or side == "RIGHT" and timerBarPos == "RIGHT")
        or (self.nr < 5 and TotemTimers_Settings.ShowCooldowns and TotemTimers_Settings.Arrange == "vertical" and self.numtimers > 1 and 
            ((side == "LEFT" and self.numtimers > 1 and TotemTimersFrame:GetLeft() > WorldFrame:GetWidth()/2)
            or (side == "RIGHT" and self.numtimers > 1 and TotemTimersFrame:GetLeft() < WorldFrame:GetWidth()/2)))
    then
		return (self.timeSpacing + self.timerbars[1]:GetWidth()*self.timerbars[1]:GetEffectiveScale())
	end
	return 0
end


function TotemTimers:SetWidth(width)
	self.button:SetWidth(width)
end

function TotemTimers:SetHeight(height)
	self.button:SetHeight(height)
end

function TotemTimers:SetFont(font)
    for _,timerbar in pairs(self.timerbars) do
        local _,height = timerbar.time:GetFont()
        timerbar.time:SetFont(font, height)
    end
end

function TotemTimers:SetTimeHeight(height)
	for _,timerbar in pairs(self.timerbars) do
		timerbar:SetHeight(height)
        local font = timerbar.time:GetFont()
		timerbar.time:SetFont(font, height)
	end
	self:Reanchor()
end

function TotemTimers:SetTimeWidth(width)
	for _,timerbar in pairs(self.timerbars) do
		timerbar:SetWidth(width)
		timerbar.time:SetWidth(width)
	end
	self:Reanchor()
end

function TotemTimers:SetScale(scale)
	self.button:SetScale(scale)
	self:SetTimerBarPos(self.timerBarPos)
	--self:Reanchor()
end

function TotemTimers:SetBarTexture(texture)
    for _,bar in pairs(self.timerbars) do
        bar:SetStatusBarTexture(texture)
        bar.background:SetStatusBarTexture(texture)
    end
end

function TotemTimers:SetBarColor(r,g,b)
    for _,bar in pairs(self.timerbars) do
        bar:SetStatusBarColor(r,g,b, 1.0)
        bar.background:SetStatusBarColor(r,g,b,0.4)
    end
end

--allowed position combinations are: CENTER/CENTER, LEFT/RIGHT, RIGHT/LEFT, TOP/BOTTOM, BOTTOM/TOP

local CounterPositions = {
	CENTER = "CENTER",
	LEFT = "RIGHT",
	RIGHT = "LEFT",
	TOP = "BOTTOM", 
	BOTTOM = "TOP",
}

local DirectionXMult = {
	CENTER = 0,
	LEFT = 1,
	RIGHT = -1,
	TOP = 0,
	BOTTOM = 0,
}

local DirectionYMult = {
	CENTER = 0,
	LEFT = 0,
	RIGHT = 0,
	TOP = -1,
	BOTTOM = -1
}


function TotemTimers:SetPoint(pos, relframe)
	local relborder = 0
	if relframe.button then 
		relborder = relframe:GetBorder(CounterPositions[pos])
		relframe = relframe.button
	end
	local border = self:GetBorder(pos)+relborder
	self.button:ClearAllPoints()
	self.button:SetPoint(pos, relframe, CounterPositions[pos], (self.spacing+border)*DirectionXMult[pos], (self.spacing+border)*DirectionYMult[pos])
end

-- anchors this timer to another
function TotemTimers:Anchor(timer, point)
	table.insert(self.anchors, {timer = timer, point = point})
	table.insert(timer.anchorchilds, self)
	self:SetPoint(point, timer) 
end

-- updates the positions of all frames anchored to this timer
function TotemTimers:Reanchor()
	for _, anchor in pairs(self.anchors) do
		self:SetPoint(anchor.point, anchor.timer)
	end
	for _, anchorchild in pairs(self.anchorchilds) do
		anchorchild:Reanchor()
	end
end

function TotemTimers:ClearAnchors()
	self.anchors = {}
	self.anchorchilds = {}
end

function TotemTimers:SetSpacing(spacing)
	self.spacing = spacing
	self:Reanchor()
end

function TotemTimers:SetTimeSpacing(spacing)
	self.timeSpacing = spacing
	self:SetTimerBarPos(self.timerBarPos)
	--self:Reanchor()
end


function TotemTimers:Show()
    self.button:Show()
    for i=1,self.numtimers do
        if self.timers[i] > 0 then
            self:ShowTimerBar(i)
        end
    end
end

function TotemTimers:Hide()
    self.button:Hide()
    for i=1,self.numtimers do
        if self.timers[i] > 0 then
            self:HideTimerBar(i)
        end
    end
end
